"""
There are several limitations to privatizing large models: high deployment costs
and poor performance.
In scenarios where data data_privacy requirements are relatively low, connecting
with commercial large models can enable
rapid and efficient product implementation with high quality.
"""
